/* implements algorithm 2.6.7 from Cohen's book (page 94).

  Copyright 2004, 2005 Vincent Lefe`vre and Paul Zimmermann.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along
  with this program; see the file COPYING.  If not, write to the Free
  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include "gmp.h"
#include "wclr.h"

/* #define DEBUG2 */

unsigned int max_size = 0;

void
init_lll (void)
{
  printf ("(using %s", GS_STRING);
#if (GS_HAS_PREC != 0)
  printf ("-%u", PRECISION);
#endif
  printf (" as Gram-Schmidt coefficients)\n");
}

/* assume b > 0 */ 
static int
mpf_cmpabs (mpf_t a, mpf_t b)
{
  if (mpf_sgn (a) > 0)
    return mpf_cmp (a, b) > 0;
  else
    {
      mpf_t t;
      int res;
      mpf_init2 (t, mpf_get_prec (a));
      mpf_abs (t, a);
      res = mpf_cmp (t, b) > 0;
      mpf_clear (t);
      return res;
    }
}

static void
reduce (mpz_t **b, GSTYPE **lam, GSTYPE *d, unsigned int k,
        unsigned int l, unsigned int m, GSTYPE q, GSTYPE tmp)
{
  unsigned int j;
#ifdef DEBUG
  static double maxq = 0.0;
#endif

#ifdef DEBUG2
  printf ("reduce %u %u\n", k, l);
#endif
#ifdef GS_FRACTION_FREE
  GS_MUL_2EXP (q, lam[k][l], 1);
  if (GS_GTABS (q, d[l]))
#else
  if (GS_GTABS_ONEHALF(lam[k][l]))
#endif
    {
      mpz_t z;
      unsigned int oldprec;

      mpz_init (z);
#ifdef GS_FRACTION_FREE
      GS_ROUND_DIV(q, q, d[l]); /* q = round(q/(2*d[l])) */
#else
      GS_ROUND(q, lam[k][l]); /* q = round(lam[k][l]) */
#endif
      MPZ_SET_GS (z, q);
      for (j = 0; j < m; j++)
        mpz_submul (b[k][j], b[l][j], z);
#ifdef GS_HAS_PREC
      oldprec = GS_GET_PREC(q);
      GS_SET_PREC(q, 1 + mpz_sizeinbase (z, 2));
#endif
#ifdef GS_FRACTION_FREE
      GS_SUBMUL (lam[k][l], q, d[l], tmp);
#else
      GS_SUB (lam[k][l], lam[k][l], q);
#endif
      for (j = 0; j < l; j++)
        GS_SUBMUL (lam[k][j], q, lam[l][j], tmp);
#ifdef GS_HAS_PREC
      GS_SET_PREC(q, oldprec);
#endif
      mpz_clear (z);
    }
}

/* B is d[k] * d[k-2] + lam[k,k-1]^2 */
static void
swap (mpz_t **b, GSTYPE **lam, GSTYPE *d, unsigned int k, unsigned int kmax,
      GSTYPE B, GSTYPE t, GSTYPE t2)
{
  unsigned int i, j;
  mpz_t *tmp;

#ifdef DEBUG2
  printf ("swap %u %u\n", k, kmax);
#endif
  if (k >= 2)
    for (j = 0; j <= k - 2; j++)
      GS_SWAP (lam[k-1][j], lam[k][j]);
#ifdef GS_FRACTION_FREE
  GS_DIVEXACT (B, B, d[k-1]);
#else 
  GS_MUL (d[k-1], d[k-1], B);
  GS_UI_DIV (B, 1, B); /* invert B to perform only one division */
  GS_MUL (t, lam[k][k-1], B);
  GS_MUL (d[k], d[k], B);
#endif
  for (i = k + 1; i <= kmax; i++)
    {
#ifdef GS_FRACTION_FREE
      /* l[i,k]   <- (d[k]*l[i,k-1] - l[k,k-1]*l[i,k])/d[k-1]
         l[i,k-1] <- (b*l[i,k] + l[k,k-1]*newl[i,k])/d[k] */
      GS_MUL (lam[i][k-1], d[k], lam[i][k-1]);
      GS_SUBMUL (lam[i][k-1], lam[k][k-1], lam[i][k], t2);
      GS_DIVEXACT (lam[i][k-1], lam[i][k-1], d[k-1]);
      GS_MUL (lam[i][k], B, lam[i][k]);
      GS_ADDMUL (lam[i][k], lam[k][k-1], lam[i][k-1], t2);
      GS_DIVEXACT (lam[i][k], lam[i][k], d[k]);
      GS_SWAP (lam[i][k-1], lam[i][k]);
#else
      GS_SUBMUL (lam[i][k-1], lam[k][k-1], lam[i][k], t2);
      GS_ADDMUL (lam[i][k], t, lam[i][k-1], t2);
      GS_SWAP (lam[i][k-1], lam[i][k]);
#endif
    }
#ifdef GS_FRACTION_FREE
  GS_SET (d[k-1], B);
#else
  GS_SET (lam[k][k-1], t);
#endif
  /* swap rows b[k] and b[k-1] */
  tmp = b[k-1];
  b[k-1] = b[k];
  b[k] = tmp;
}

/* integral lll on B[0]...B[n-1], each having m coefficients
   Lovasz parameters a/b.
   Returns number of calls to reduce or swap.
*/
unsigned long
lllint (mpz_t **B, unsigned int n, unsigned int m)
{
  unsigned int k, kmax, i, j;
  unsigned long calls = 0;
  GSTYPE *d, **lam, u, v, t;
  mpz_t w;
#ifdef DEBUG
  static unsigned int msize = 0;
  int *perm;
  perm = (int*) malloc (n * sizeof(int));
  if (perm == NULL)
    {
      fprintf (stderr, "malloc failed\n");
      exit (1);
    }
  for (i = 0; i < n; i++)
    perm[i] = i;
#endif

#ifdef DEBUG
  for (i = 0; i < n; i++)
    for (j = 0; j < m; j++)
      {
        k = mpz_sizeinbase (B[i][j], 2);
        if (k > msize)
          printf ("matrix size=%u (%u)\n", msize = k, n);
      }
#endif

  k = 1;
  kmax = 0;

  INIT (u);
  INIT (v);
  INIT (t);
  mpz_init (w);
  d = (GSTYPE*) malloc (n * sizeof (GSTYPE));
  if (d == NULL)
    {
      fprintf (stderr, "malloc failed\n");
      exit (1);
    }
  for (i = 0; i < n; i++)
    INIT (d[i]);
  GS_SET_Z(v, B[0][0]);
  GS_MUL(d[0], v, v);
  for (j = 1; j < m; j++)
    {
      GS_SET_Z(v, B[0][j]);
      GS_ADDMUL(d[0], v, v, t);
    }
 /* the Gram-Schmidt coefficients mu[i,j] = scalmul(b[i],bstar[j]) / B[j]
     satisfy mu[i,j]=lam[i,j]/d[j], where both lam[i,j] and d[j] are integers,
     and d[i] := B[1]*...*B[i], where B[i] := scalmul(bstar[i], bstar[i]).
     Floating-point case: take lam[i,j]=scalmul(b[i],bstar[j]), d[i]=B[i].
  */
  lam = (GSTYPE**) malloc (n * sizeof (GSTYPE*));
  if (lam == NULL)
    {
      fprintf (stderr, "malloc failed\n");
      exit (1);
    }
  for (i = 0; i < n; i++)
    {
      lam[i] = (GSTYPE*) malloc (n * sizeof (GSTYPE));
      if (lam[i] == NULL)
        {
          fprintf (stderr, "malloc failed\n");
          exit (1);
        }
      for (j = 0; j < n; j++)
        INIT (lam[i][j]);
    }
  do
    {
      if (k > kmax)
        {
          kmax = k;
          for (j = 0; j <= k; j++)
            {
              GS_SET_Z (v, B[k][0]);
              GS_SET_Z (t, B[j][0]);
              GS_MUL (u, v, t);
              for (i = 1; i < m; i++)
                {
                  GS_SET_Z (v, B[k][i]);
                  GS_SET_Z (t, B[j][i]);
                  GS_ADDMUL (u, v, t, t);
                }
              for (i = 0; i < j; i++)
                {
#ifdef GS_FRACTION_FREE
                  GS_MUL (u, u, d[i]);
                  GS_SUBMUL (u, lam[k][i], lam[j][i], v);
                  if (i > 0)
                    GS_DIVEXACT (u, u, d[i-1]);
#else
                  GS_MUL (v, lam[k][i], lam[j][i]);
                  GS_SUBMUL (u, d[i], v, t);
#endif
                }
              if (j < k)
                {
#ifdef GS_FRACTION_FREE
                  GS_SET (lam[k][j], u);
#else
                  GS_DIV (lam[k][j], u, d[j]);
#endif
                }
              else
                {
                  if (GS_ISZERO(u))
                    {
                      fprintf (stderr, "Error, not a basis\n");
                      exit (1);
                    }
                  GS_SET (d[k], u);
                }
            }
        }
      do
        {
          reduce (B, lam, d, k, k - 1, m, u, v);
          calls ++;
#ifdef GS_FRACTION_FREE
          if (k == 1)
            GS_SET (u, d[k]);
          else
            GS_MUL (u, d[k], d[k-2]);
#else
          GS_DIV (u, d[k], d[k-1]);
#endif
          GS_ADDMUL (u, lam[k][k-1], lam[k][k-1], v);
#ifdef GS_FRACTION_FREE
          GS_MUL (v, d[k-1], d[k-1]);
          GS_MUL_A (v, v);
          GS_DIV_B (v, v); /* divide v by LLL_B */
          if (GS_LT(u, v))
            {
#else
              GS_MUL_B (u, u);
          if (GS_LT_UI(u, LLL_A))
            {
              GS_DIV_B (u, u); /* restore u */
#endif
#ifdef DEBUG
              i = perm[k];
              perm[k] = perm[k-1];
              perm[k-1] = i;
#endif
              swap (B, lam, d, k, kmax, u, v, t);
              calls ++;
              k = (k > 1) ? k - 1 : 1;
            }
          else
            break;
        }
      while (1);
      for (i = k - 1; i > 0;)
        {
          reduce (B, lam, d, k, --i, m, u, v);
          calls ++;
        }
      k ++;
    }
  while (k < n);

  for (i = 0; i < n; i++)
    {
      for (j = 0; j < n; j++)
        CLEAR (lam[i][j]);
      free (lam[i]);
    }
  free (lam);
  for (i = 0; i < n; i++)
    CLEAR (d[i]);
  free (d);
  CLEAR (u);
  CLEAR (v);
  CLEAR (t);
  mpz_clear (w);
#ifdef DEBUG
  /*  for (i = 0; i < n; i++)
    printf ("%u ", perm[i]);
    printf ("\n"); */
  free (perm);
#endif
  return calls;
}

/* same as lllint, but cut coefficients by k bits.
 Assumes n=m=5, and the matrix B is lower triangular. */
unsigned long
lllint2 (mpz_t **B, unsigned int n, unsigned int m, unsigned int k)
{
  mpz_t **C;
  mpz_t t;
  unsigned int i, j, r;
  unsigned long muls;
  static unsigned int perm[5] = {1, 2, 0, 3, 4};

  if (n != 5 || m != 5)
    abort ();

  C = malloc (n * sizeof (mpz_t*));
  if (C == NULL)
    {
      fprintf (stderr, "malloc failed\n");
      exit (1);
    }
  for (i = 0; i < n; i++)
    {
      C[i] = malloc (m * sizeof (mpz_t));
      if (C[i] == NULL)
        {
          fprintf (stderr, "malloc failed\n");
          exit (1);
        }
      for (j = 0; j < m; j++)
        {
          mpz_init (C[i][j]);
          mpz_div_2exp (C[i][j], B[i][perm[j]], k);
        }
    }
  
  mpz_init (t);
  /* reduce C */
  muls = lllint (C, n, m);

  for (i = 0; i < n; i++)
    {
      /* compute l[0..4] such that C[i] = l[0]*B[0]+...+l[4]*B[4]
         put l[0..4] directly in C[i][0..4] */
      for (j = m; j > 0; )
        {
          j--;
          for (r = j + 1; r < m; r++)
            {
              mpz_div_2exp (t, B[r][perm[j]], k);
              mpz_submul (C[i][j], t, C[i][r]);
            }
          mpz_div_2exp (t, B[j][perm[j]], k);
          mpz_divexact (C[i][j], C[i][j], t);
        }
    }

  /* now multiply C * B */
  for (i = 0; i < n; i++) /* compute C[i] * B */
    for (j = 0; j < m; j++) /* compute coefficient j */
      {
        mpz_set_ui (t, 0);
        for (r = j; r < m; r++)
          mpz_addmul (t, C[i][r], B[r][perm[j]]);
        mpz_set (C[i][j], t);
      }

  /* copy C into B and clear C */
  for (i = 0; i < n; i++)
    {
      for (j = 0; j < m; j++)
        {
          mpz_set (B[i][perm[j]], C[i][j]);
          mpz_clear (C[i][j]);
        }
      free (C[i]);
    }
  free (C);
  mpz_clear (t);
  return muls;
}

static void
Print_mat (mpz_t **B, int rows, int cols) /* prints to NTL format */
{
  long i, j; 

  printf ("[");
  for (i = 0; i < rows; i++)
    {
      printf ("[");
      for (j = 0; j < cols; j++)
        {
          mpz_out_str (stdout, 10, B[i][j]);
          if (j < cols - 1)
            printf (" "); 
        }
      printf ("]\n"); 
    }
  printf ("]\n");
}

void
Print_maple (mpz_t **B, int rows, int cols) /* prints to Maple format */
{
  long i, j; 

  printf ("[");
  for (i = 0; i < rows; i++)
    {
      printf ("[");
      for (j = 0; j < cols; j++)
        {
          mpz_out_str (stdout, 10, B[i][j]);
          if (j < cols - 1)
            printf (",");
        }
      printf ("]");
      if (i < rows - 1)
        printf (",");
      printf ("\n");
    }
  printf ("]:\n");
}

/* copy B[0..n-1][0..m-1] into B[1..n][1..m] */
void
post_copy (mpz_t **B, unsigned long n, unsigned long m)
{
  unsigned long i, j;

  for (i = n; i >= 1; i--)
    for (j = m; j >= 1; j--)
      mpz_set (B[i][j], B[i-1][j-1]);
}

#ifdef MAIN
int
main (int argc, char *argv[])
{
  unsigned int n, m, i, j;
  mpz_t **B;
  int c;

  if (argc != 3)
    {
      fprintf (stderr, "Usage: lllint <n> <m> < file\n");
      exit (1);
    }
  
  n = atoi(argv[1]);
  m = atoi(argv[2]);
  B = (mpz_t**) malloc(n * sizeof(mpz_t*));
  if (B == NULL)
    {
      fprintf (stderr, "malloc failed\n");
      exit (1);
    }
  c = getchar(); while (isspace(c) || c=='\n') c=getchar();
  if (c != '[') { fprintf(stderr, "Error: '[' expected\n"); exit(1); }
  for (i=0; i<n; i++)
    {
      B[i] = (mpz_t*) malloc(m * sizeof(mpz_t));
      if (B[i] == NULL)
        {
          fprintf (stderr, "malloc failed\n");
          exit (1);
        }
      for (j=0; j<m; j++)
        {
          mpz_init (B[i][j]);
        }
    }
  fprintf(stderr, "Considering each row as a vector.\n");
  for (i=0; i<n; i++)
    {
      c = getchar();
      while (isspace(c) || c=='\n')
        c = getchar ();
      if (c != '[')
        {
          fprintf (stderr, "Error at row %d: '[' expected instead of %c\n",
                   i, c);
          exit (1);
        }
      for (j=0; j<m; j++)
        mpz_inp_str(B[i][j], stdin, 0);
      c = getchar();
      while (isspace(c) || c=='\n')
        c = getchar ();
      if (c != ']')
        {
          fprintf (stderr, "Error: ']' expected at line %u\n", i);
          exit (1);
        }
    }
  c = getchar();
  while (isspace(c) || c=='\n')
    c=getchar ();
  if (c != ']')
    {
      fprintf (stderr, "Error: ']' expected\n");
      exit (1);
    }

  lllint (B, n, m);
  printf ("Reduced matrix is:\n");
  Print_mat (B, n, m);
  for (i=0; i<n; i++)
    {
      for (j=0; j<m; j++)
        mpz_clear (B[i][j]);
      free(B[i]);
    }
  free(B);

  return 0;
}
#endif
